# NumGfun version 0.1
#
# Recurrence unrolling by binary splitting; analytic continuation of holonomic functions; bounds for
# holonomic sequences.
#
# Marc Mezzarobba, MPRI master internship, projet ALGO, INRIA Rocquencourt, february-july 2007

# Procedures marked (*) are highly experimental, inefficient, or lack some cases.
#
#
# ANALYTIC CONTINUATION
#
#   ac_eval(deq, y(z), path, precision)
#       Numerical analytic continuation along path of the solution of deq (in 0).
#       deq - differential equation with initial values
#       y(z) - variables
#       path - list of elements of Q[i] starting in 0 (e.g. [0,1,1+I])
#       precision - (absolute) number of digits needed
#       
#   transition_matrix(deq, y(z), path, precision)
#       Same as ac_eval, but compute the whole transition matrix.
#       deq - differential equation (initial values are ignored)
#       y(z), path, precision - as above
#       
#   diffeqtoproc
#
#   local_monodromy(deq, y(z), sing, start, precision)  (*)
#       Trivial wrapper around transition_matrix. 
#       start - basepoint
#       sing - singularity
#       deq, y(z), precision - as above
#   
# NTH TERM FOR RECURRENT SEQUENCES
#
#   nth_term(rec, u(n), N [, ringname])
#       Nth term of the sequence defined by rec.
#       deq - recurrence equation (with initial values)
#       u(n) - variables
#       N - positive integer
#       ringname - if set to ndmatrix, compute Sum(u(j),j=0..n-1) instead
#
#   fnth_term
#       Same as nth_term, with floating-point instead of rational result
#       
#   rectoproc_binsplit(rec, u(n))
#       Returns a procedure computing u as defined by rec.
#       rec, u(n) - as above
#
# BOUNDS
#   
#   bound_diffeq(deq,y(z))
#       Compute a majorant series valid for all fundamental solutions of deq.
#       deq - differential equation (initial values are ignored)
#       y(z) - variables
#       
#   bound_rec(rec, u(n))
#       Compute a simple bound for the solution of rec.
#       rec - recurrence equation (with initial values)
#       u(n) - variables
#       
#   bound_ratpoly(rat,z)
#       Computes a sharp (exact up to logarithmic factors) majorant series for rat.
#       rat - rational polynomial
#       z - variable
#
# MISC
#       
#   waksman_product(r)
#       Returns a procedure to multiply r*r matrices using Waksman's algorithm.


# USERINFO
#
#   <10    normal
#   10-19  verbose
#   >= 20  verbose+slow
#
#infolevel[ancont] := 10;
#infolevel[bounds] := 10;
#infolevel[nth_term] := 10;
#infolevel[step_transition_matrix] := 1;
#infolevel[binsplit] := 10;
#infolevel[recmatrix] := 10;

# TODO:
# - meilleurs types pour les récurrences

####################################################################################################
## Tuning and customization
####################################################################################################

macro(

    # - le scindage binaire gagne largement, mettre 10000 ci-dessous pour s'en convaincre
    # - calibré pour ndmatrix ou ndseries
    binary_splitting_threshold = 800,
    
    # return precision_factor*N digits when actually computing N (useful to test sharp bounds)
    precision_factor = 1,
    
    # when N terms would seem necessary, compute terms_factor*N+terms_delta instead
    terms_factor = 1,
    terms_delta = 0,

    #binsplit = `binsplit/inlined`,
    #binsplit = `binsplit/generic`,

    none = none # final ','

);

####################################################################################################
## Module definition
####################################################################################################

NumGfun := module()

description
    "Binary splitting, analytic continuation of holonomic functions, and bounds";

option
    package,
    load = alpha_warning;

global
    `convert/ndmatrix`,
    `type/ndmatrix`,
    `type/matrix_ring`,
    `value/Coeftayl`,
    ndmatrix,
    
    opaquemodules,
    # used by LinearAlgebra
    shape, identity, zero, outputoptions, inplace,
    
    # module names for userinfo
    ancont, bounds; 

export 
    ac_eval,
    bound_diffeq,
    bound_rec,
    bound_ratpoly,
    diffeqtoproc,
    local_monodromy,
    ndseries,
    nth_term,
    fnth_term,
    rectoproc_binsplit,
    transition_matrix,
    monodromy,
    waksman_product,
    NumGfun_version;

local
    # matrix rings
    LAMain, ReIm, bigint, mvMult, WaksmanInplace, waksman,
    # misc names
    generic, nbterms,

    fail_if_singular,
    evalf_absoluteprec,
    `Nth_term/common`,
    `Nth_term/ndseries`,
    `nth_term/doit`,
    `binsplit/inlined`,
    `makeitfloat/Q[i]`,
    `makeitfloat/ndmatrix`,
    `needed_terms/sharp`,
    `needed_terms/vdH01/zero`,
    `needed_terms/vdH01/maj`,
    `rest_majorant/vdHlike`,
    alpha_warning,
    binsplit,
    bound_ratpoly_coeffs,
    bound_ratpoly_int,
    bound_ratpoly_on_disk,
    bound_transition_matrix,
    bound_values,
    call,
    dominant_root,
    extractline,
    first_root,
    frectomatrix,
    frectopoly,
    genmatring,
    getname,
    deq_inicond_vector,
    majorant_series,
    make_fresh_name_generator,
    rec_inicond_vector,
    makeitfloat,
    myformatrec,
    ndmatrix_multiply,
    ndmatrix_norm,
    needed_terms,
    old_needed_terms,
    orddiffeq,
    ordfrec,
    ordrec,
    parametered_rec,
    path_transition_matrix,
    pole_multiplicity,
    read_diffeq,
    read_diffeq_or_rec,
    read_rec,
    recmatrix,
    rectodiffrec,
    root_multiplicity,
    singularities,
    solve_ineq,
    step_transition_matrix,
    vdH01_bound,
    
    matrix_ring,
    s_opaquemodules;

alpha_warning := proc()
    WARNING(cat(
        "NumGfun is pre-alpha quality code, don't trust it! Please report bugs to marc",
        "@", "mezzarobba", ".", "net."));
end proc;

NumGfun_version := "0.1";

use gfun, LinearAlgebra in

#binsplit := `binsplit/generic`;

s_opaquemodules := kernelopts(opaquemodules);
kernelopts(opaquemodules=false);

$include "utilities.mpl"
$include "matrices.mpl"
$include "nthterm.mpl"
$include "bounds.mpl"
$include "ancont.mpl"
$include "diffeqtoproc.mpl"

kernelopts(opaquemodules=s_opaquemodules);

end use;

end module;

savelib(
    NumGfun,
    `convert/ndmatrix`,
    `type/ndmatrix`,
    `type/matrix_ring`,
    `value/Coeftayl`
);
